﻿using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using VL.Core;
using VL.Lib.Collections;

namespace VL.Dijkstra
{
    [Type]
    public class TraversalData
    {
        [Node]
        public ImmutableDictionary<Vertex, TraversalEntry> Data { get; private set; }

        [Node]
        public Spread<Vertex> All { get; private set; }

        [Node]
        public Vertex Target { get; private set; }

        [Node]
        public TraversalData(IEnumerable<Vertex> vertices)
        {
            Target = null;
            Data = ImmutableDictionary.Create<Vertex, TraversalEntry>();
            All = Spread.Create<Vertex>();

            Reset(vertices);
        }

        [Node]
        public void Reset(IEnumerable<Vertex> vertices)
        {
            Target = null;
            Data = Data.Clear();

            All = vertices.ToSpread();
        }

        [Node]
        public void FindPathes(Vertex targetVertex)
        {
            if (!All.Contains(targetVertex))
            {
                targetVertex = null;
            }

            Target = targetVertex;
            Data = Data.Clear();

            var Unvisited = All.ToList();

            var newData = new Dictionary<Vertex, TraversalEntry>();
            foreach (var v in All)
            {
                newData[v] = new TraversalEntry();
            }


            var targetInfo = newData[targetVertex];
            targetInfo.DistanceToTarget = 0;

            while (targetVertex != null)
            {
                var visited = VisitVertex(targetVertex, newData);
                Unvisited.Remove(visited);

                targetVertex = Unvisited.Where(v => !double.IsPositiveInfinity(newData[v].DistanceToTarget))
                                        .OrderBy(v => newData[v].DistanceToTarget)
                                        .FirstOrDefault();
            }

            Data = ImmutableDictionary.ToImmutableDictionary(newData);

        }


        [Node]
        public IEnumerable<int> GetPath(Vertex fromVertex)
        {
            if (Target != null)
            {
                yield return fromVertex.Index;
                while (Data[fromVertex].PreviousVertex != null)
                {
                    fromVertex = (Vertex)Data[fromVertex].PreviousVertex;
                    yield return fromVertex.Index;
                }
            }
        }

        [Node]
        public override string ToString()
        {
            var anon = this;
            return "Vertices: " + All.Count() + "\nEdges: " + All.Select(v => v.Neighbours.Count).Sum() + "\nLongestPath: " + All.Select(v => anon.GetPath(v).Count()).Max() ;
        }


        //        [Node]
        private Vertex VisitVertex(Vertex vertex, IReadOnlyDictionary<Vertex, TraversalEntry> data)
        {
            var distance = data[vertex].DistanceToTarget;

            foreach (var n in vertex.Neighbours.Keys)
            {
                var d = distance + vertex.Neighbours[n];

                var entry = data[n];

                if (entry.DistanceToTarget > d)
                {
                    entry.DistanceToTarget = d;
                    entry.PreviousVertex = vertex;
                }
            }

            return vertex;
        }
    }
}
